@@ -1,9 +1,9 @@
 =begin
-  top-level extconf.rb for Ruby-GNOME2
+  top-level extconf.rb for Ruby-GNOME
 
   $Id: extconf.rb,v 1.17 2007/10/22 12:19:17 ktou Exp $
 
-  Copyright (C) 2003-2005 Ruby-GNOME2 Project Team
+  Copyright (C) 2003-2025  Ruby-GNOME Project Team
 =end
 
 require 'English'
@@ -11,16 +11,20 @@ require 'mkmf'
 require 'fileutils'
 require 'pathname'
 
-priorlibs = [
+prior_libs = [
   "glib2",
+  "cairo-gobject",
   "gobject-introspection",
   "gio2",
   "gdk_pixbuf2",
   "pango",
   "atk",
-  "gtk2",
   "gdk3",
   "gtk3",
+  "graphene1",
+  "gdk4",
+  "gsk4",
+  "gtk4",
 ]
 
 unsupported_libraries = [
@@ -49,15 +53,15 @@ $topdir = File.expand_path($topdir)
 
 subdirs = ARGV.select{|v|  /^--/ !~ v}
 
-if subdirs.size == 0
+if subdirs.empty?
   subdirs = Dir.glob($topsrcdir+"/*/extconf.rb")
   subdirs.collect! do |subdir|
     subdir[0..$topsrcdir.size] = ""
     File.dirname(subdir)
   end
-  priorlibs &= subdirs
-  subdirs -= priorlibs
-  subdirs = priorlibs + subdirs #Change the order
+  prior_libs &= subdirs
+  subdirs -= prior_libs
+  subdirs = prior_libs + subdirs # Change the order
 end
 subdirs = subdirs.delete_if do |subdir|
   subdir.end_with?("-no-gi")
@@ -78,6 +82,7 @@ else
 end
 
 subdirs.each do |subdir|
+  puts "::group::Run extconf.rb: #{subdir}"
   STDERR.puts("#{$0}: Entering directory `#{subdir}'")
   FileUtils.mkdir_p(subdir)
   topdir = File.join(*([".."] * subdir.split(/\/+/).size))
@@ -94,8 +99,11 @@ subdirs.each do |subdir|
   else
     ignore_modules << subdir
   end
+  puts "::endgroup::"
 end
-puts "\n-----"
+
+puts "::group::Show top-level extconf.rb results"
+puts "-----"
 unless target_modules.empty?
   puts "Target libraries: #{target_modules.join(', ')}"
 end
@@ -147,5 +155,6 @@ end
 
 puts "-----"
 puts "Done."
+puts "::endgroup::"
 
 $makefile_created = true
