@@ -1,4 +1,4 @@
-# Copyright (C) 2012-2017  Ruby-GNOME2 Project Team
+# Copyright (C) 2012-2025  Ruby-GNOME Project Team
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -24,6 +24,9 @@ module GObjectIntrospection
         loader.version = options[:version]
         loader.load(namespace)
       end
+
+      def initialize_instance_post(object)
+      end
     end
 
     attr_accessor :version
@@ -35,14 +38,16 @@ module GObjectIntrospection
     def load(namespace)
       repository = Repository.default
       repository.require(namespace, @version)
-      pre_load(repository, namespace)
-      repository.each(namespace) do |info|
-        load_info(info) if info.is_a?(InterfaceInfo)
+      prepare_class(@base_module) do
+        pre_load(repository, namespace)
+        repository.each(namespace) do |info|
+          load_info(info) if info.is_a?(InterfaceInfo)
+        end
+        repository.each(namespace) do |info|
+          load_info(info) unless info.is_a?(InterfaceInfo)
+        end
+        post_load(repository, namespace)
       end
-      repository.each(namespace) do |info|
-        load_info(info) unless info.is_a?(InterfaceInfo)
-      end
-      post_load(repository, namespace)
     end
 
     private
@@ -77,39 +82,44 @@ module GObjectIntrospection
 
     def load_function_info(info)
       name = rubyish_method_name(info)
-      define_singleton_method(@base_module, name, info)
+      define_singleton_method(info, @base_module, name)
+      define_equal_style_setter(info, @base_module.singleton_class, name)
     end
 
-    def define_module_function(target_module, name, function_info)
-      function_info.unlock_gvl = should_unlock_gvl?(function_info, target_module)
-      method_name = "#{target_module}\#.#{name}"
-      arguments_builder = ArgumentsBuilder.new(function_info, method_name)
-      target_module.module_eval do
-        define_method(name) do |*arguments, &block|
-          arguments_builder.build(arguments, block)
-          arguments = arguments_builder.arguments
-          block = arguments_builder.block
-          function_info.invoke(arguments, &block)
-        end
-        module_function(name)
-      end
+    def load_function_info_singleton_method(info, klass, method_name)
+      define_singleton_method(info, klass, method_name)
+      define_equal_style_setter(info, klass.singleton_class, method_name)
     end
 
-    def define_singleton_method(klass, name, info)
-      info.unlock_gvl = should_unlock_gvl?(info, klass)
-      arguments_builder = ArgumentsBuilder.new(info, "#{klass}.#{name}")
-      require_callback_p = info.require_callback?
-      singleton_class = (class << klass; self; end)
-      singleton_class.__send__(:define_method, name) do |*arguments, &block|
-        arguments_builder.build(arguments, block)
-        arguments = arguments_builder.arguments
-        block = arguments_builder.block
-        if block.nil? and require_callback_p
-          to_enum(name, *arguments)
-        else
-          info.invoke(arguments, &block)
+    def define_module_function(function_info, target_module, name)
+      prepare_function_info_lock_gvl(function_info, target_module)
+      full_method_name = "#{target_module}\#.#{name}"
+      invoker = Invoker.new(function_info, name, full_method_name)
+      target_module::INVOKERS[name] = invoker
+      target_module.module_eval(<<-DEFINE_METHOD, __FILE__, __LINE__ + 1)
+        def #{name}(*arguments, &block)
+          INVOKERS["#{name}"].invoke(nil, arguments, block)
         end
+        module_function(:#{name})
+      DEFINE_METHOD
+    end
+
+    def define_singleton_method(info, klass, name)
+      case name
+      when "hash"
+        # Ignore methods that require 1+ arguments and use well-known
+        # name. It may break something.
+        return unless info.n_required_in_args.zero?
       end
+      prepare_function_info_lock_gvl(info, klass)
+      invoker = Invoker.new(info, name, "#{klass}.#{name}")
+      singleton_class = klass.singleton_class
+      singleton_class::INVOKERS[name] = invoker
+      singleton_class.class_eval(<<-DEFINE_METHOD, __FILE__, __LINE__ + 1)
+        def #{name}(*arguments, &block)
+          INVOKERS["#{name}"].invoke(nil, arguments, block)
+        end
+      DEFINE_METHOD
     end
 
     def define_struct(info, options={})
@@ -124,8 +134,10 @@ module GObjectIntrospection
                                         :parent => options[:parent],
                                         :size   => size)
       end
-      load_fields(info, klass)
-      load_methods(info, klass)
+      prepare_class(klass) do
+        load_fields(info, klass)
+        load_methods(info, klass)
+      end
     end
 
     def load_struct_info(info)
@@ -137,7 +149,13 @@ module GObjectIntrospection
         end
         unless method_infos.empty?
           base_class = @base_module.const_get(base_class_name)
-          load_method_infos(method_infos, base_class.singleton_class)
+          singleton_class = base_class.singleton_class
+          invokers = singleton_class.const_get(:INVOKERS).dup
+          singleton_class.__send__(:remove_const, :INVOKERS)
+          singleton_class.const_set(:INVOKERS, invokers)
+          load_methods_method(method_infos, singleton_class)
+          Ractor.make_shareable(singleton_class::INVOKERS) if defined?(Ractor)
+          singleton_class.private_constant(:INVOKERS)
         end
       else
         return if info.gtype_struct?
@@ -147,6 +165,7 @@ module GObjectIntrospection
     end
 
     def define_boxed(info)
+      return if info.gtype == GLib::Type::NONE
       klass = self.class.define_class(info.gtype, info.name, @base_module)
       _ = klass # TODO: Remove me. It is just for suppressing a warning.
       # TODO
@@ -159,13 +178,20 @@ module GObjectIntrospection
     end
 
     def load_enum_value(value_info, enum_module)
-      enum_module.const_set(value_info.name.upcase, value_info.value)
+      value = value_info.value
+      Ractor.make_shareable(value) if defined?(Ractor)
+      enum_module.const_set(value_info.name.upcase, value)
     end
 
     def define_enum(info)
-      self.class.define_class(info.gtype,
-                              rubyish_class_name(info),
-                              @base_module)
+      # TODO: Can we do the same things for flags on NONE GType?
+      return if info.gtype == GLib::Type::NONE
+      klass = self.class.define_class(info.gtype,
+                                      rubyish_class_name(info),
+                                      @base_module)
+      prepare_class(klass) do
+        load_methods(info, klass)
+      end
     end
 
     def define_error(info)
@@ -208,7 +234,12 @@ module GObjectIntrospection
         end
         @base_module.const_set(rubyish_class_name(info), flags_module)
       else
-        self.class.define_class(info.gtype, flags_class_name(info), @base_module)
+        klass = self.class.define_class(info.gtype,
+                                        flags_class_name(info),
+                                        @base_module)
+        prepare_class(klass) do
+          load_methods(info, klass)
+        end
       end
     end
 
@@ -217,11 +248,17 @@ module GObjectIntrospection
     end
 
     def load_object_info(info)
+      return if info.gtype == GLib::Type::NONE
+      # e.g.: GstBitmask
+      return if info.fundamental? and !info.gtype.instantiatable?
       klass = self.class.define_class(info.gtype,
                                       rubyish_class_name(info),
                                       @base_module)
-      load_fields(info, klass)
-      load_methods(info, klass)
+      prepare_class(klass) do
+        load_virtual_functions(info, klass)
+        load_fields(info, klass)
+        load_methods(info, klass)
+      end
     end
 
     def load_fields(info, klass)
@@ -285,11 +322,11 @@ module GObjectIntrospection
         next if method_infos.empty?
         case method_infos.first
         when ConstructorInfo
-          load_constructor_infos(method_infos, klass)
+          load_methods_constructor(method_infos, klass)
         when MethodInfo
-          load_method_infos(method_infos, klass)
+          load_methods_method(method_infos, klass)
         when FunctionInfo
-          load_function_infos(method_infos, klass)
+          load_methods_function(method_infos, klass)
         else
           raise "TODO: #{method_info_class}"
         end
@@ -300,133 +337,64 @@ module GObjectIntrospection
       end
     end
 
-    def load_constructor_infos(infos, klass)
+    def load_methods_constructor(infos, klass)
       return if infos.empty?
 
-      call_initialize_post = lambda do |object|
-        initialize_post(object)
-      end
+      klass.const_set(:LOADER_CLASS, self.class)
+      invokers = {}
+      klass.const_set(:INITIALIZE_INVOKERS, invokers)
       infos.each do |info|
         name = "initialize_#{info.name}"
-        info.unlock_gvl = should_unlock_gvl?(info, klass)
-        arguments_builder = ArgumentsBuilder.new(info, "#{klass}\##{name}")
-        klass.__send__(:define_method, name) do |*arguments, &block|
-          arguments_builder.build(arguments, block)
-          arguments = arguments_builder.arguments
-          block = arguments_builder.block
-          info.invoke(self, arguments, &block)
-          call_initialize_post.call(self)
-        end
-        klass.__send__(:private, name)
-      end
-
-      find_info = lambda do |arguments|
-        find_suitable_callable_info(infos, arguments)
-      end
-      klass.__send__(:define_method, "initialize") do |*arguments, &block|
-        info = find_info.call(arguments, &block)
-        __send__("initialize_#{info.name}", *arguments, &block)
-      end
-    end
-
-    def initialize_post(object)
-    end
-
-    def find_suitable_callable_info(infos, arguments)
-      min_n_args = nil
-      max_n_args = nil
-      candidate_infos = []
-      infos.each do |info|
-        n_in_args = info.n_in_args
-        n_required_in_args = info.n_required_in_args
-        if (n_required_in_args..n_in_args).cover?(arguments.size)
-          candidate_infos << info
-        end
-        min_n_args = [min_n_args || n_required_in_args, n_required_in_args].min
-        max_n_args = [max_n_args || n_in_args, n_in_args].max
-      end
-
-      if candidate_infos.size == 1
-        return candidate_infos.first
-      elsif candidate_infos.size > 1
-        candidate_info = candidate_infos.find do |info|
-          in_arg_infos = info.in_args
-          arguments.each.with_index.all? do |argument, i|
-            match_argument?(in_arg_infos[i], argument)
+        prepare_function_info_lock_gvl(info, klass)
+        invoker = Invoker.new(info, name, "#{klass}\##{name}")
+        invokers[name] = invoker
+        klass.class_eval(<<-DEFINE_METHOD, __FILE__, __LINE__ + 1)
+          def #{name}(*arguments, &block)
+            invoker = INITIALIZE_INVOKERS["#{name}"]
+            invoker.invoke(self, arguments, block)
+            LOADER_CLASS.initialize_instance_post(self)
           end
-        end
-        return candidate_info || candidate_infos.first
+          private :#{name}
+        DEFINE_METHOD
       end
 
-      detail = "#{arguments.size} for #{min_n_args}"
-      if min_n_args < max_n_args
-        detail << "..#{max_n_args}"
-      end
-      raise ArgumentError, "wrong number of arguments (#{detail})"
+      klass.class_eval(<<-DEFINE_METHOD, __FILE__, __LINE__ + 1)
+        def initialize(*arguments, &block)
+          invokers = INITIALIZE_INVOKERS
+          invokers.values.each do |invoker|
+            catch do |tag|
+              invoker.invoke(self, arguments.dup, block, tag)
+              LOADER_CLASS.initialize_instance_post(self)
+              return
+            end
+          end
+          full_method_name = "\#{self.class.name}\#\#{__method__}"
+          message = +"wrong arguments: "
+          message << "\#{full_method_name}("
+          message << arguments.collect(&:inspect).join(", ")
+          message << "):\\n"
+          message << "available signatures:"
+          invokers.each_value do |invoker|
+            message << "\\n\#{full_method_name}\#{invoker.signature}"
+          end
+          raise ArgumentError, message
+        end
+      DEFINE_METHOD
+      Ractor.make_shareable(klass::INITIALIZE_INVOKERS) if defined?(Ractor)
     end
 
-    def match_argument?(arg_info, argument)
-      type = arg_info.type
-      case type.tag
-      when TypeTag::BOOLEAN
-        argument == true or argument == false
-      when TypeTag::INT8,
-           TypeTag::UINT8,
-           TypeTag::INT16,
-           TypeTag::UINT16,
-           TypeTag::INT32,
-           TypeTag::UINT32,
-           TypeTag::INT64,
-           TypeTag::UINT64,
-           TypeTag::FLOAT,
-           TypeTag::DOUBLE
-        argument.is_a?(Numeric)
-      when TypeTag::GTYPE
-        argument.is_a?(GLib::Type)
-      when TypeTag::UTF8
-        argument.is_a?(String)
-      when TypeTag::FILENAME
-        argument.is_a?(String)
-      when TypeTag::ARRAY
-        argument.is_a?(Array)
-      when TypeTag::INTERFACE
-        interface = type.interface
-        case interface.type
-        when InfoType::STRUCT
-          match_argument_interface_struct?(arg_info, interface, argument)
-        when InfoType::OBJECT,
-             InfoType::INTERFACE,
-             InfoType::FLAGS,
-             InfoType::ENUM
-          argument.is_a?(interface.gtype.to_class)
-        else
-          # TODO
-          false
-        end
-      when TypeTag::GLIST,
-           TypeTag::GSLIST
-        argument.is_a?(Array)
-      else
-        # TODO
-        false
-      end
+    def load_virtual_functions(info, klass)
+      klass.extend(VirtualFunctionImplementable)
+      gtype_prefix = rubyish_gtype_name(klass.gtype.name)
+      implementor = VirtualFunctionImplementor.new(self.class,
+                                                   gtype_prefix,
+                                                   info.vfuncs)
+      klass.__send__(:initialize_virtual_function_implementable,
+                     implementor)
     end
 
-    def match_argument_interface_struct?(arg_info, interface, argument)
-      gtype = interface.gtype
-      case gtype.name
-      when "void"
-        # TODO
-        false
-      when "CairoSurface"
-        if Object.const_defined?(:Cairo)
-          argument.is_a?(Cairo::Surface)
-        else
-          false
-        end
-      else
-        argument.is_a?(gtype.to_class)
-      end
+    def rubyish_gtype_name(name)
+      name.scan(/[A-Z]+[a-z\d]+/).collect(&:downcase).join("_")
     end
 
     def rubyish_method_name(function_info, options={})
@@ -461,6 +429,36 @@ module GObjectIntrospection
           else
             name
           end
+        when "not_equal"
+          if n_in_args == 1
+            "!="
+          else
+            name
+          end
+        when "less_than"
+          if n_in_args == 1
+            "<"
+          else
+            name
+          end
+        when "less_than_or_equal"
+          if n_in_args == 1
+            "<="
+          else
+            name
+          end
+        when "greater_than"
+          if n_in_args == 1
+            ">"
+          else
+            name
+          end
+        when "greater_than_or_equal"
+          if n_in_args == 1
+            ">="
+          else
+            name
+          end
         when /\A(?:is|get_is|can_be)_/
           "#{$POSTMATCH}?"
         when /\Aget_/
@@ -500,7 +498,13 @@ module GObjectIntrospection
             name
           end
         when "to_string"
-          "to_s"
+          if n_in_args.zero?
+            "to_s"
+          else
+            name
+          end
+        when "to_integer"
+          "to_i"
         when "foreach"
           "each"
         else
@@ -539,11 +543,15 @@ module GObjectIntrospection
       end
     end
 
-    def should_unlock_gvl?(function_info, klass)
-      false
+    def prepare_function_info_lock_gvl(function_info, target_module)
+      # For backward compatibility
+      if respond_to?(:should_unlock_gvl?)
+        function_info.lock_gvl_default =
+          !should_unlock_gvl?(function_info, target_module)
+      end
     end
 
-    def load_method_infos(infos, klass)
+    def load_methods_method(infos, klass)
       infos.each do |info|
         method_name = rubyish_method_name(info)
         load_method_info(info, klass, method_name)
@@ -563,36 +571,22 @@ module GObjectIntrospection
     end
 
     def define_method(info, klass, method_name)
-      info.unlock_gvl = should_unlock_gvl?(info, klass)
+      return if method_name.empty?
+      prepare_function_info_lock_gvl(info, klass)
       remove_existing_method(klass, method_name)
-      function_info_p = (info.class == FunctionInfo)
-      have_return_value_p = info.have_return_value?
-      arguments_builder = ArgumentsBuilder.new(info, "#{klass}\##{method_name}")
-      require_callback_p = info.require_callback?
-      klass.__send__(:define_method, method_name) do |*arguments, &block|
-        arguments.unshift(self) if function_info_p
-        arguments_builder.build(arguments, block)
-        arguments = arguments_builder.arguments
-        block = arguments_builder.block
-        if block.nil? and require_callback_p
-          to_enum(method_name, *arguments)
-        else
-          if function_info_p
-            return_value = info.invoke(arguments, &block)
-          else
-            return_value = info.invoke(self, arguments, &block)
-          end
-          if have_return_value_p
-            return_value
-          else
-            self
-          end
+      invoker = Invoker.new(info, method_name, "#{klass}\##{method_name}")
+      invokers = klass::INVOKERS
+      invokers[method_name] = invoker
+      klass.class_eval(<<-DEFINE_METHOD, __FILE__, __LINE__ + 1)
+        def #{method_name}(*arguments, &block)
+          invoker = INVOKERS["#{method_name}"]
+          invoker.invoke(self, arguments, block)
         end
-      end
+      DEFINE_METHOD
     end
 
     def define_equal_style_setter(info, klass, method_name)
-      if /\Aset_/ =~ method_name and info.n_args == 1
+      if /\Aset_/ =~ method_name and info.n_in_args == 1
         setter_method_name = "#{$POSTMATCH}="
         remove_existing_method(klass, setter_method_name)
         klass.__send__(:alias_method, setter_method_name, method_name)
@@ -601,27 +595,33 @@ module GObjectIntrospection
 
     def define_inspect(info, klass, method_name)
       if method_name == "to_s" and info.n_args.zero?
-        klass.__send__(:define_method, "inspect") do ||
-          super().gsub(/\>\z/) {" #{to_s}>"}
-        end
+        klass.class_eval(<<-DEFINE_METHOD, __FILE__, __LINE__ + 1)
+          def inspect
+            super.gsub(/>\\z/) {" \#{to_s}>"}
+          end
+        DEFINE_METHOD
       end
     end
 
-    def load_function_infos(infos, klass)
+    def load_methods_function(infos, klass)
       infos.each do |info|
         name = rubyish_method_name(info)
         next if name == "new"
         next if name == "alloc"
-        define_singleton_method(klass, name, info)
+        load_function_info_singleton_method(info, klass, name)
       end
     end
 
     def load_interface_info(info)
+      return if info.gtype == GLib::Type::NONE
       interface_module =
         self.class.define_interface(info.gtype,
                                     rubyish_class_name(info),
                                     @base_module)
-      load_methods(info, interface_module)
+      prepare_class(interface_module) do
+        load_virtual_functions(info, interface_module)
+        load_methods(info, interface_module)
+      end
     end
 
     def load_constant_info(info)
@@ -633,36 +633,103 @@ module GObjectIntrospection
     end
 
     def load_union_info(info)
+      return if info.gtype == GLib::Type::NONE
       klass = self.class.define_class(info.gtype, info.name, @base_module)
-      load_fields(info, klass)
-      load_methods(info, klass)
+      prepare_class(klass) do
+        load_fields(info, klass)
+        load_methods(info, klass)
+      end
     end
 
-    class ArgumentsBuilder
-      attr_reader :arguments
-      attr_reader :block
-      def initialize(info, method_name)
+    def prepare_class(klass)
+      pre_prepare_class(klass)
+      yield
+      post_prepare_class(klass)
+    end
+
+    def pre_prepare_class(klass)
+      klass.const_set(:INVOKERS, {})
+      klass.singleton_class.const_set(:INVOKERS, {})
+    end
+
+    def post_prepare_class(klass)
+      return unless defined?(Ractor)
+      Ractor.make_shareable(klass::INVOKERS)
+      Ractor.make_shareable(klass.singleton_class::INVOKERS)
+      klass.private_constant(:INVOKERS)
+      klass.singleton_class.private_constant(:INVOKERS)
+    end
+
+    def define_methods_module(name)
+      mod = Module.new
+      @base_module.const_set(name, mod)
+      mod.const_set(:INVOKERS, {})
+      mod
+    end
+
+    def apply_methods_module(mod, target)
+      target.include(mod)
+      post_methods_module(mod)
+    end
+
+    def post_methods_module(mod)
+      return unless defined?(Ractor)
+      Ractor.make_shareable(mod::INVOKERS)
+      mod.private_constant(:INVOKERS)
+    end
+
+    class Invoker
+      def initialize(info, method_name, full_method_name)
         @info = info
         @method_name = method_name
-
+        @full_method_name = full_method_name
         @prepared = false
-
-        @arguments = nil
-        @block = nil
+        ensure_prepared if defined?(Ractor)
       end
 
-      def build(arguments, block)
-        prepare
+      def invoke(receiver, arguments, block, abort_tag=nil)
+        ensure_prepared
 
-        @arguments = arguments
-        @block = block
+        if receiver and @function_info_p
+          arguments.unshift(receiver)
+        end
 
-        build_arguments
-        validate_arguments
+        arguments, block = build(receiver, arguments, block)
+        if wrong_number_of_arguments?(arguments)
+          if abort_tag
+            throw(abort_tag)
+          else
+            raise ArgumentError, invalid_error_message(arguments)
+          end
+        end
+        unless normalize_arguments!(arguments, abort_tag)
+          return @value_on_invalid
+        end
+
+        if block.nil? and @require_callback_p
+          receiver.to_enum(@method_name, *arguments)
+        else
+          if @function_info_p
+            return_value = @info.invoke(arguments, &block)
+          else
+            return_value = @info.invoke(receiver, arguments, &block)
+          end
+          if @have_return_value_p
+            return_value
+          else
+            receiver
+          end
+        end
+      end
+
+      def signature
+        ensure_prepared
+        argument_signatures = @in_args.collect(&:signature)
+        "(" + argument_signatures.join(", ") + ")"
       end
 
       private
-      def prepare
+      def ensure_prepared
         return if @prepared
 
         @in_args = @info.in_args
@@ -676,40 +743,134 @@ module GObjectIntrospection
         end
         @valid_n_args_range = (@n_required_in_args..@n_in_args)
 
+        @in_arg_types = []
+        @in_arg_nils = []
+        @in_arg_nil_indexes = []
+        @in_args.each_with_index do |arg, i|
+          @in_arg_types << arg.type
+          may_be_null = arg.may_be_null?
+          @in_arg_nils << may_be_null
+          @in_arg_nil_indexes << i if may_be_null
+        end
+
+        @function_info_p = (@info.class == FunctionInfo)
+        @have_return_value_p = @info.have_return_value?
+        @require_callback_p = @info.require_callback?
+
+        prepare_on_invalid
+
         @prepared = true
       end
 
-      def build_arguments
-        if @block and @last_in_arg_is_gclosure
-          @arguments << @block
-          @block = nil
-        end
-
-        n_missing_arguments = @n_in_args - @arguments.size
-        if n_missing_arguments > 0
-          nil_indexes = []
-          @in_args.each_with_index do |arg, i|
-            nil_indexes << i if arg.may_be_null?
-          end
-          return if nil_indexes.size < n_missing_arguments
-          nil_indexes.each_with_index do |nil_index, i|
-            next if i <= n_missing_arguments
-            @arguments.insert(nil_index, nil)
-          end
+      def prepare_on_invalid
+        case @method_name
+        when "=="
+          @value_on_invalid = false
+          @on_invalid = :fallback
+        when "!="
+          @value_on_invalid = true
+          @on_invalid = :fallback
+        else
+          @on_invalid = :raise
         end
       end
 
-      def validate_arguments
-        return if @valid_n_args_range.cover?(@arguments.size)
+      def build(receiver, arguments, block)
+        if block and @last_in_arg_is_gclosure
+          arguments << block
+          block = nil
+        end
 
-        detail = "#{@arguments.size} for "
+        n_missing_arguments = @n_in_args - arguments.size
+        n_in_arg_nil_indexes = @in_arg_nil_indexes.size
+        if 0 < n_missing_arguments and n_missing_arguments < n_in_arg_nil_indexes
+          @in_arg_nil_indexes[-n_missing_arguments..-1].each do |nil_index|
+            arguments.insert(nil_index, nil)
+          end
+        end
+
+        return arguments, block
+      end
+
+      def wrong_number_of_arguments?(arguments)
+        not @valid_n_args_range.cover?(arguments.size)
+      end
+
+      def normalize_arguments!(arguments, abort_tag)
+        arguments.size.times do |i|
+          argument = arguments[i]
+          type = @in_arg_types[i]
+          converted_argument = type.try_convert(argument)
+          if converted_argument.nil?
+            next if argument.nil? and @in_arg_nils[i]
+            if abort_tag
+              throw(abort_tag)
+            elsif @on_invalid == :fallback
+              return false
+            else
+              next
+            end
+          end
+          arguments[i] = converted_argument
+        end
+        true
+      end
+
+      def invalid_error_message(arguments)
+        detail = "#{arguments.size} for "
         if @n_in_args == @n_required_in_args
           detail << "#{@n_in_args}"
         else
           detail << "#{@n_required_in_args}..#{@n_in_args}"
         end
-        message = "#{@method_name}: wrong number of arguments (#{detail})"
-        raise ArgumentError, message
+        "#{@full_method_name}: wrong number of arguments (#{detail})"
+      end
+    end
+
+    class VirtualFunctionImplementor
+      def initialize(loader_class, gtype_prefix, infos)
+        @loader_class = loader_class
+        @gtype_prefix = gtype_prefix
+        @infos = {}
+        prefix = GLib::VIRTUAL_FUNCTION_IMPLEMENTATION_PREFIX
+        infos.each do |info|
+          name = info.name
+          @infos[:"#{prefix}#{name}"] = info
+          @infos[:"#{prefix}#{gtype_prefix}_#{name}"] = info
+        end
+      end
+
+      def implement(implementor_gtype, name)
+        info = @infos[name]
+        return false if info.nil?
+        container = info.container
+        vtable_gtype = container.gtype
+        if container.respond_to?(:class_struct)
+          struct = container.class_struct
+        else
+          return false unless implementor_gtype.type_is_a?(vtable_gtype)
+          struct = container.iface_struct
+        end
+        field = struct.find_field(info.name)
+        @loader_class.implement_virtual_function(field,
+                                                 implementor_gtype,
+                                                 vtable_gtype,
+                                                 name.to_s)
+        true
+      end
+    end
+
+    module VirtualFunctionImplementable
+      def initialize_virtual_function_implementable(implementor)
+        @virtual_function_implementor = implementor
+      end
+
+      def implement_virtual_function(implementor_class, name)
+        unless instance_variable_defined?(:@virtual_function_implementor)
+          return false
+        end
+        @virtual_function_implementor.implement(implementor_class.gtype,
+                                                name)
       end
     end
   end
